/*
 * Copyright (c) 2004
 * Telecooperation Office (TecO), Universitaet Karlsruhe (TH), Germany.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the Universitaet Karlsruhe (TH) nor the names
 *    of its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * @author Philipp Scholl <scholl@teco.edu>
 */

package edu.teco.particles.pComponent;

import java.util.Iterator;

import javax.swing.JScrollPane;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeCellRenderer;

import edu.teco.particles.libParticle.*;
import edu.teco.particles.pComponent.io.ParticlePacketListener;
import edu.teco.particles.pComponent.io.ParticlePacketsEvent;
import edu.teco.particles.pComponent.networkMap.*;

/**
 * Maps the network into a browsable tree.
 * 
 * Functionality: 
 * - tree view of the network. 
 * - browse network by: - bridge -> particle -> tuple - particle -> tuple -> position in packet 
 * - showed will beall information which can be devised from a particlePacket 
 * - if a packet from the same srcid is received, the old one will be replaced by the new one.
 */
public class NetworkMap extends JScrollPane implements ParticlePacketListener
{

  /**
   * creates the NetworkMap component.
   */
  public NetworkMap()
  {
    setViewportView(getView());
  }

  /** the interface for rec'ving packets from the network. */
  public void particlePackets(ParticlePacketsEvent e)
  {
    assert (_socket != null);

    // loop over all packets adding them to the model.
    synchronized (_socket.getBuffer())
    {
      for (Iterator i = _socket.getBuffer().iterator(); i.hasNext();)
      {
        ParticlePacket packet;
        packet = (ParticlePacket) i.next();

        MutableTreeNode root = (MutableTreeNode) getModel().getRoot();

        if (!NodeFactory.contains(packet))
        {
          getModel().insertNodeInto(NodeFactory.get(packet), root, 0);
        }
      }
    }
  }

  /**
   * @name MVC part.
   * @short defines factory methods for the view and model. creates default
   *        model and view for this component. In case you want to use a
   *        customized model or view with this controller you would need to
   * implement set methods here. @{
   */
  protected NetworkMapModel _model;

  public NetworkMapModel getModel()
  {
    if (_model == null)
    {
      _model = new NetworkMapModel();
    }
    return _model;
  }

  protected NetworkMapView _view;

  public NetworkMapView getView()
  {
    if (_view == null)
    {
      _view = new NetworkMapView();
      _view.setModel(getModel());
      _view.setCellRenderer(getCellRenderer());
    }
    return _view;
  }

  protected TreeCellRenderer _renderer;

  public TreeCellRenderer getCellRenderer()
  {
    if (_renderer == null)
    {
      _renderer = new CellRenderer();
    }
    return _renderer;
  }

  /** @} */

  /**
   * @name SocketThreadProperty
   * @short Property for the SocketThread from which packets are received. @{
   */
  private SocketThread _socket;

  public SocketThread getSocket()
  {
    return _socket;
  }

  public void setSocket(SocketThread socket)
  {
    assert (socket != null);
    _socket = socket;
    socket.addParticlePacketListener(this);
  }
  /** @} */
}
