/*
 * Created on 16.12.2004
 *
 * Copyright (c) 2004
 * Telecooperation Office (TecO), Universitaet Karlsruhe (TH), Germany.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the Universitaet Karlsruhe (TH) nor the names
 *    of its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 *  @author scholl@teco.edu
 */
package edu.teco.particles.pComponent;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.TreePath;

import edu.teco.particles.libParticle.ParticlePacket;
import edu.teco.particles.libParticle.ParticleTuple;
import edu.teco.particles.pComponent.io.ParticlePacketListener;
import edu.teco.particles.pComponent.io.ParticlePacketsEvent;
import edu.teco.particles.pComponent.scope.Chart;
import edu.teco.particles.pComponent.scope.ContextEditor;
import edu.teco.particles.pComponent.scope.Trace;
import edu.teco.particles.pComponent.util.ParticleContext;
import edu.teco.particles.pComponent.util.TupleParser;
import edu.teco.particles.pComponent.util.TupleParserFactory;

/**
 * Facade of Scope Particle Component.
 * 
 * Functionality: - decodes data to signed/unsigned byte/long at offset. - shows
 * only a fixed window size of data - slides when the window is full. - foreach
 * acl type there is one trace.
 * 
 * @author Philipp Scholl <scholl@teco.edu>
 */
public class Scope extends JSplitPane implements ParticlePacketListener,
    TreeSelectionListener
{

  /**
   * holds the startTime of this Object. needed for offset calculation
   */
  private long _startTime = System.currentTimeMillis();

  /**
   * creates a new Scope.
   */
  public Scope()
  {
    super();

    setOrientation(JSplitPane.VERTICAL_SPLIT);
    setLeftComponent(_getChart());
    setRightComponent(_getConfPanel());
    setResizeWeight(0.9D);
  }

  /**
   * creates a new Container for the Configuration Component.
   * 
   * @return a Container.
   */
  private Container _getConfPanel()
  {
    BorderLayout layout = new BorderLayout();
    return new JPanel(layout);
  }

  public void particlePackets(ParticlePacketsEvent e)
  {
    assert (_socket != null);
    addPackets(_socket.getBuffer(), _socket.getTimeStamps());
  }

  /** reference for the Chart */
  protected Chart _chart;

  /** factory method */
  protected Chart _getChart()
  {
    if (_chart == null)
    {
      _chart = new Chart();
    }

    return _chart;
  }

  /**
   * calculates the timeoffsets from timeStamps and _startTime.
   * 
   * @param timeStamps
   *          a List of timeStamps to consider
   * @return the timeOffsets of the timeStamps
   */
  private long[] _getTimeOffsets(List<Date> timeStamps)
  {
    long timeOffsets[];

    timeOffsets = new long[timeStamps.size()];
    Date date;

    for (int i = 0; i < timeStamps.size(); i++)
    {
      date = timeStamps.get(i);
      timeOffsets[i] = (date.getTime() - _startTime);
      // if( i != 0 && timeOffsets[i-1] == timeOffsets[i] ) {
      // System.out.println( "offset" + (double) timeOffsets[i] );
      // }
    }

    return timeOffsets;
  }

  /**
   */
  public void addPackets(List<ParticlePacket> buffer, List<Date> timeStamps)
  {
    long timeOffsets[] = _getTimeOffsets(timeStamps);

    // iterate over the packets + timeOffsets
    Iterator<ParticlePacket> bufferIt = buffer.iterator();
    ParticlePacket packet;
    ParticleTuple tuple;
    ParticleContext context;
    Trace trace;
    String tupleType;
    Double data;
    TupleParser parser;

    // iterate over all packets and timeStamps
    int i = 0;
    context = new ParticleContext();

    while (bufferIt.hasNext() && (timeOffsets[i] != 0))
    {
      packet = bufferIt.next();
      tuple = packet.firstAcl();
      context.setSrcId(packet.getSrcId());

      // iterate over tuples and parse them
      try
      {
        while (tuple != null)
        {
          context.setTuple(tuple);
          trace = _getChart().getTrace(context);

          if (trace != null)
          {
            parser = TupleParserFactory.getTupleParser(context);
            data = ((Number) parser.parse(tuple)).doubleValue();

            trace.addPoint((double) timeOffsets[i], data);
          }

          tuple = packet.nextAcl(tuple);
        }
      }
      catch (IllegalArgumentException e)
      {
        // hmm, seems that a wrong parser was chosen
        // we'll just ignore this !
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
      i++;
    }

    _getChart().repaint();
  }

  /**
   * clears the Chart.
   */
  public void clear()
  {
    _getChart().clear();
    _startTime = System.currentTimeMillis();
  }

  /**
   * @name SocketProperty
   * @brief The Socket from which packets are recvd. @{
   */
  private SocketThread _socket;

  public SocketThread getSocket()
  {
    return _socket;
  }

  public void setSocket(SocketThread socket)
  {
    _socket = socket;
    socket.addParticlePacketListener(this);
  }

  /** @} */

  /**
   * @name ResolutionProperty
   * @brief defines the Resolution Property. The Resolution of the Scope is the
   *        interval in milliseconds between subsequent sample points of the
   * trace. So setting the Reolution to 13ms draws a point every 13ms. @{
   */
  public double getResolution()
  {
    return _getChart().getResolution();
  }

  public void setResolution(double res)
  {
    _getChart().setResolution(res);
  }

  /** @} */

  /**
   * @name drawImpulses
   * @brief determines wether the Chart should be drawn with Impulses or not. @{
   */
  public boolean isDrawImpulses()
  {
    return _getChart().isDrawImpulses();
  }

  public void setDrawImpulses(boolean state)
  {
    _getChart().setDrawImpulses(state);
  }

  /** @} */

  /**
   * @name maxY
   * @brief defines the maximum value on the Y axis @{
   */
  public int getMaxY()
  {
    return _getChart().getMaxY();
  }

  public void setMaxY(int max)
  {
    _getChart().setMaxY(max);
  }

  /** @} */

  /**
   * @name MinY
   * @brief defines the Minimum value on the Y axis @{
   */
  public int getMinY()
  {
    return _getChart().getMinY();
  }

  public void setMinY(int Min)
  {
    _getChart().setMinY(Min);
  }

  /** @} */

  /**
   * @name DrawGrid
   * @brief defines wether the unit Grid should be drawn or not @{
   */
  public boolean isDrawXGrid()
  {
    return _getChart().isDrawXGrid();
  }

  public void setDrawXGrid(boolean state)
  {
    _getChart().setDrawXGrid(state);
  }

  /** @} */

  /**
   * @name DrawGrid
   * @brief defines wether the unit Grid should be drawn or not @{
   */
  public boolean isDrawYGrid()
  {
    return _getChart().isDrawYGrid();
  }

  public void setDrawYGrid(boolean state)
  {
    _getChart().setDrawYGrid(state);
  }

  /** @} */

  /**
   * @name drawContinious
   * @brief determines wether the Chart should be drawn with connected lines. @{
   */
  public boolean isDrawContinious()
  {
    return _getChart().isDrawContinious();
  }

  public void setDrawContinious(boolean state)
  {
    _getChart().setDrawContinious(state);
  }

  /** @} */

  /** reference for the ContextTreeRenderer */
  protected ContextEditor _ctxEditor;

  /**
   * create a Component for a ParticleContext.
   * 
   * @param ctx
   *          the context to create the configuration dialog for.
   * @return a component that can be used to configurate a context.
   */
  protected Component _getConfiguration(ParticleContext ctx)
  {
    if (_ctxEditor == null)
    {
      _ctxEditor = new ContextEditor(ctx);
    }

    _ctxEditor.setContext(ctx);

    return _ctxEditor;
  }

  /*
   * (non-Javadoc)
   * 
   * @see javax.swing.event.TreeSelectionListener#valueChanged(javax.swing.event.TreeSelectionEvent)
   */
  public void valueChanged(TreeSelectionEvent e)
  {
    ParticleContext context = null;

    // walk through the whole tree and see what's selected and what's not.
    for (TreePath path : e.getPaths())
    {
      try
      {
        context = (ParticleContext) path.getLastPathComponent();

        if (e.isAddedPath(path))
        {
          _getChart().add(context);
          setRightComponent(_getConfiguration(context));
        }
        else
        {
          _getChart().remove(context);
        }
      }
      catch (ClassCastException ee)
      {
        // do nothing;
      }
    }
  }

}
