/*
 * Copyright (c) 2004
 * Telecooperation Office (TecO), Universitaet Karlsruhe (TH), Germany.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the Universitaet Karlsruhe (TH) nor the names
 *    of its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
package edu.teco.particles.pComponent;

import java.util.*;
import java.util.concurrent.CopyOnWriteArrayList;

import edu.teco.particles.libParticle.*;
import edu.teco.particles.pComponent.io.*;
import edu.teco.particles.pComponent.util.TupleParser;
import edu.teco.particles.pComponent.util.TupleParserFactory;

/**
 * a threaded ParticleSocket Bean.
 * 
 * @author scholl <scholl@teco.edu>
 * 
 * ParticleSocket that can be used as a Thread in your Application.
 * 
 * This is a Wrapper for ParticleSocket. It implements the Observer pattern for
 * communicating recv'd packets to other components.
 * 
 * Note: The Thread is polling a the ParticleSocket, which has some Implications
 * on the timestamping mechanism for packets, which don't have a CTS Tuple. The
 * Maximum Resolution for this kind of timestamping is _POLLDELAY ! this might
 * also cause trouble for relative CTS parsing.
 * 
 * Further Note: why use polling instead of blocking mode: - since this app is
 * interactive and you can't interrupt the Thread that is currently blocked,
 * there is no _clean_ solution to update for example the currently active
 * filter while not receiving anything.
 */
public class SocketThread extends Thread
{
  /**
   * the socket to listen for ParticlePackets.
   */
  private ParticleSocket _recvSocket;

  /**
   * the socket to send on
   */
  private ParticleSocket _sendSocket;

  public SocketThread()
  {
    super();

    _recvSocket = _createParticleSocket(_recvPort);
    _sendSocket = _createParticleSocket(_sendPort);
  }

  /**
   * factory method for a ParticleSocket some circumstances can cause that a
   * socket can't be created at the current time, in this case we try until a
   * socket is available.
   */
  protected ParticleSocket _createParticleSocket(int port)
  {
    ParticleSocket sock = null;
    boolean created = false;

    while (!created)
    {
      try
      {
        sock = new ParticleSocket(port);
        sock.setOptions(P_SOCKET_OPTIONS.SOCKET_FRAG
            | P_SOCKET_OPTIONS.SOCKET_RECV_ALL, 1);
        sock.setOptions(P_SOCKET_OPTIONS.SOCKET_BLOCKING, 0);

        created = true;
      }
      catch (UnknownError e)
      {
        // hmm, currently the socket can't be created -> try again
      }
    }

    assert (sock != null);
    return sock;
  }

  /**
   * Not the _buffer and _timeStamps are CopyOnWriteLists !!
   */
  private List<ParticlePacket> _buffer     = new CopyOnWriteArrayList<ParticlePacket>();

  private List<Date>           _timeStamps = new CopyOnWriteArrayList<Date>();

  /*
   * ===========================================================================
   * the Thread part of this Component
   * ========================================================================
   */
  private volatile Thread      _thread;

  // the minimum network packet
  private static int           _POLLDELAY  = 10;

  public void run()
  {
    try
    {
      ParticlePacket packet = null;
      long timeWaiting = 0;

      while (true)
      {
        // wait until we are not stopped
        synchronized (this)
        {
          while (_threadStopped)
            wait();
        }

        assert (_filter != null);
        assert (_sendSocket != null);
        assert (_recvSocket != null);

        synchronized (this)
        {
          while ((packet = _recvSocket.receiveFiltered(_sendSocket, _filter)) != null)
          {
            _buffer.add(packet);
            _timeStamps.add(_getTimeStamp(packet));
          }
        }

        if (timeWaiting > _updateDelay && _buffer.size() != 0)
        {
          _notifyParticlePacketObservers();
          _buffer.clear();
          _timeStamps.clear();
          timeWaiting = 0;
        }
        else
        {
          timeWaiting += _POLLDELAY;
        }

        synchronized (this)
        {
          wait(_POLLDELAY);
        }
      }
    }
    catch (InterruptedException e)
    {
      e.printStackTrace();
    }
  }

  private synchronized void _startReceiving()
  {
    if (_thread == null)
    {
      _thread = new Thread(this);
      _thread.start();
      _thread.setName("socketThread");
    }
    _threadStopped = false;
    notify();
  }

  private synchronized void _stopReceiving()
  {
    _threadStopped = true;
  }

  /**
   * return a list of received Packets. The Buffer is cleared after each
   * _updateDelay.
   */
  public List<ParticlePacket> getBuffer()
  {
    return _buffer;
  }

  /**
   * returns an array of TimeStamps. The Receive Time of the Packet with the
   * same Index as the buffer returned by getBuffer().
   */
  public List<Date> getTimeStamps()
  {
    return _timeStamps;
  }

  /**
   * send a particlePacket over the the _sendSocket. delegate to _sendSocket.
   * 
   * @param pck
   *        the packet to be send.
   */
  public synchronized void send(ParticlePacket pck)
  {
    _sendSocket.send(pck);
  }

  /** delegate to _sendSocket */
  public synchronized void sendAcked(ParticlePacket pck, ParticleSrcId dest)
  {
    assert (pck != null);
    assert (dest != null);
    _sendSocket.sendAcked(_recvSocket, pck, dest);
  }

  /** deleaget to _sendSocket */
  public synchronized void sendAdressed(ParticlePacket pck, ParticleSrcId dest)
  {
    assert (pck != null);
    assert (dest != null);
    _sendSocket.sendAddressed(pck, dest);
  }

  /** delegate to _sendSocket */
  public synchronized void setRetry(short retries, long timeout)
  {
    _sendSocket.setRetry(retries, timeout);
  }

  /** delegate to _sendSocket */
  public synchronized void setSrcId(ParticleSrcId id)
  {
    _sendSocket.setId(id);
  }

  /**
   * creates a timeStamp for the Packet. NOTE: CTS is only decoded if _decodeCTS
   * is true.
   * 
   * @param pck
   *          the packet to create the timeStamp for
   * @return returns the CTS coded TimeStamp or the current time
   */
  protected Date _getTimeStamp(ParticlePacket pck)
  {
    Date curr = new Date(System.currentTimeMillis());

    try
    {
      ParticleTuple ctsTuple = pck.findFirstAcl("cts");
      if (_decodeCts)
      {
        Date date;
        date = (Date) _timeParser.parse(ctsTuple);
        date = _tryRelativeCTSHack(pck, date);
        return date;
      }
      else
      {
        return curr;
      }
    }
    catch (IllegalArgumentException e)
    {
      // the tuple wasn't found in the packet
      return curr;
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }

    return null;
  }

  /**
   * some cts tuple define the time relative to their "power-on" time.
   * 
   * So we try make at least the difference between the packets exact. We do
   * this by capturing the first timeStamp received from a given srcId.
   * Afterwards we add all cts-data to this first timeStamp.
   * 
   * The first TimeStamp is created by the factory as the current time.
   * 
   * If the first TimeStamp is not == 0, the factory subtracts the parsed date
   * from the current time.
   * 
   * @param pck
   *          the packet to create the timestamp for.
   * @param date
   *          the parsed CTS-Data.
   * @return a difference correct TimeStamp.
   */
  protected Date _tryRelativeCTSHack(ParticlePacket pck, Date date)
  {
    Calendar cal;
    Date firstStamp; // the firstStamp received for the context, not this is
    // absolute
    Date timeStamp; // and this is relative.

    cal = Calendar.getInstance();
    cal.setTime(date);

    firstStamp = DateFactory.getTimeStamp(pck.getSrcId(), date);

    if (cal.get(Calendar.YEAR) == 1970)
    {
      timeStamp = new Date(firstStamp.getTime() + date.getTime());
    }
    else
    {
      timeStamp = date;
    }

    assert (timeStamp.after(firstStamp));
    assert (timeStamp.after(date));
    return timeStamp;

  }

  /** holds a CTS parser */
  protected TupleParser _timeParser    = TupleParserFactory.create("cts");

  /*
   * ===========================================================================
   * defining the Properties of this bean
   * ========================================================================
   */

  /**
   * starts/stops the thread.
   */
  private boolean       _threadStopped = true;

  public synchronized void setReceiving(boolean state)
  {
    if (state == true)
    {
      _startReceiving();
    }
    else
    {
      _stopReceiving();
    }
  }

  public boolean isReceiving()
  {
    return !_threadStopped;
  }

  /**
   * the delay to tell the observers that new packets arrived.
   */
  private int _updateDelay = 200;

  public synchronized void setUpdateDelay(int delay) 
  throws IllegalArgumentException
  {
    if (delay<0)
      throw new IllegalArgumentException("Update Delay can't be negative.");
    _updateDelay = delay;
    notify();
  }

  public int getUpdateDelay()
  {
    return _updateDelay;
  }

  /**
   * the port to listen on.
   */
  private int _recvPort = 5555;

  public synchronized void setRecvPort(int port)
  {
    _recvPort = port;
    _recvSocket = _createParticleSocket(_recvPort);
  }

  public int getRecvPort()
  {
    return _recvPort;
  }

  /**
   * the port to send on.
   */
  private int _sendPort = 5556;

  public synchronized void setSendPort(int port)
  {
    _sendPort = port;
    _sendSocket = _createParticleSocket(_sendPort);
  }

  public int getSendPort()
  {
    return _sendPort;
  }

  /**
   * @name FilterProperty
   * @short the filter for the receive Operation. @{
   */
  private ParticleFilter _filter = new ParticleFilter("");

  public synchronized void setFilter(ParticleFilter filter)
  {
    assert (filter != null) : "argh filter is a null Object";
    _filter = filter;
  }

  public ParticleFilter getFilter()
  {
    return _filter;
  }

  /** @} */

  /**
   * @name decodeCtsProperty
   * @short defines wether CTS tuple are regarded for timeStamp Calculation or
   *        not. There are two supported types of CTS tuples. The first is that
   *        the CTS tuple is absolute, so defining the exact time when the
   *        packet was sent. Second, relative timing to the switch on time of
   * the particle. @{
   */
  protected boolean _decodeCts = false;

  public synchronized void setDecodeCTS(boolean flag)
  {
    _decodeCts = flag;
  }

  public boolean getDecodeCTS()
  {
    return _decodeCts;
  }

  /** @} */

  /*
   * ===========================================================================
   * defining Observer Helper Functions
   * ========================================================================
   */
  private List<ParticlePacketListener> _particlePacketListeners = new ArrayList<ParticlePacketListener>();

  public synchronized void addParticlePacketListener(ParticlePacketListener l)
  {
    _particlePacketListeners.add(l);
  }

  public synchronized void removeParticlePacketListener(ParticlePacketListener l)
  {
    _particlePacketListeners.remove(l);
  }

  private int openend = 0;

  private synchronized void _notifyParticlePacketObservers()
  {
    ParticlePacketsEvent event = new ParticlePacketsEvent(this);
    for (ParticlePacketListener l : _particlePacketListeners)
    {
      openend++;
      l.particlePackets(event);
      openend--;

      if (openend != 0)
      {
        // we are in major trouble!
        // Some of the listening aren't synchronized anymore.
        // be sure to surround every access to the _buffer or timestamps
        // with a synchronized( buffer ) { vlalalla....; } !
        System.out.println("UH OH");
      }
    }

    notify();
  }

  /**
   * delegate to one of the ParticleSocket.
   * 
   * @return the srcid of the socket.
   */
  public ParticleSrcId getSrcId()
  {
    // TODO socket.getSrcId() verwenden !
    return new ParticleSrcId("1.0.0.0.127.0.0.1");
  }
}
