/*
 * Copyright (c) 2004
 * Telecooperation Office (TecO), Universitaet Karlsruhe (TH), Germany.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the Universitaet Karlsruhe (TH) nor the names
 *    of its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package edu.teco.particles.pComponent.io;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import edu.teco.particles.libParticle.ParticlePacket;
import edu.teco.particles.libParticle.ParticleTuple;
import edu.teco.particles.pComponent.util.TupleParser;
import edu.teco.particles.pComponent.util.TupleParserFactory;

/**
 * A Template for tuples.
 * 
 * Functionality: - regexp like matching for tuple data.
 * 
 * Examples: new TupleTemplate("aps"); // matches all tuples of type aps new
 * TupleTemplate("aps", "0 0"); // matches all tuples of type aps with payload 0
 * 0
 */
public class TupleTemplate implements ParticleTemplate
{

  protected ParticleTuple _tuple;

  protected int           _occurences;

  protected String        _payload;

  /**
   * Creates a Template that only matches the type of a Tuple.
   * 
   * @param type
   *          the 3 letter type to match against.
   * @throws IllegalArgumentException
   *           if the type is not a valid tuple type. For example longer than 3
   *           chars.
   */
  public TupleTemplate(String type) throws IllegalArgumentException
  {
    _tuple = new ParticleTuple(type);
    _occurences = 0;
    _payload = "";
  }

  /**
   * Creates a Template that matches the tuple type and the payload.
   * 
   * @param type
   *          the type to match against.
   * @param payload
   * @throws IllegalArgumentException
   */
  public TupleTemplate(String type, String payload)
      throws IllegalArgumentException
  {
    _tuple = new ParticleTuple(type);
    _occurences = 0;
    _payload = payload;

    // oh this looks verquert but uses a side effect of setParser.
    setParser(getParser());
  }

  /**
   * a property to match the number of occurences of a tuple in the packet.
   * 
   * Imagine a Packet with two Tuple, both of type SGX. to reference the first
   * of both you set the Number of this tuple to 0. to reference the second you
   * set the Number to 1.
   * 
   * If you set the number to null, the number will not be regarded in the
   * matching.
   * 
   * @param number
   *          the number of a tuple.
   */
  public void setOccurences(int number)
  {
    _occurences = number;
  }

  /**
   * matches this template against a tuple.
   */
  public boolean match(ParticlePacket packet)
  {
    if (getTuple(packet, _occurences, _payload) != null)
    {
      _lastPacket = packet;
      notifyActionListeners();
      return true;
    }
    return false;
  }

  /**
   * finds the tuple matching this template.
   * 
   * @param packet
   *          where to look?
   * @param n
   *          which occurence is to be used? n=0 is a special case, which
   *          ignores the occurence parameter.
   * @param payload
   *          the payload to match.
   * 
   * @return null if no tuple was found.
   */
  protected ParticleTuple getTuple(ParticlePacket packet, int n, String payload)
  {
    assert (packet != null);
    ParticleTuple tuple = packet.findFirstAcl(_tuple.getAclType());
    int i = 0;
    assert (_tuple.getAclType() != null);

    while (tuple != null)
    {
      if (_payloadMatches(tuple, payload) && i == n)
        return tuple;
      else
        tuple = packet.findNextAcl(tuple);

      if (n != 0) i++;
    }

    return null;
  }

  /**
   * match a tuple payload against a spec.
   * 
   * @param tuple
   *          the tuple to match
   * @param payload
   *          the payload it should match
   * 
   * @return true if the payload in the tuple matches the specified payload.
   */
  protected boolean _payloadMatches(ParticleTuple tuple, String payload)
  {
    return payload.equals("") ? true : tuple.toString().startsWith(payload);
  }

  protected TupleParser _parser = null;

  public TupleParser getParser()
  {
    if (_parser == null)
      _parser = TupleParserFactory.create(TupleParserFactory.SHORT);
    return _parser;
  }

  public void setParser(TupleParser parser)
  {
    _parser = parser;

    if (!_payload.equals("")) _parser.setOffset(_payload.split(" ").length);
  }

  /**
   * parses a given tuple.
   * 
   * deserialazation from a ParticleTuple.
   */
  public Object parse(ParticlePacket packet) throws IllegalArgumentException
  {
    ParticleTuple tuple = getTuple(packet, _occurences, _payload);
    return getParser().parse(tuple);
  }

  protected ParticlePacket _lastPacket;

  /**
   * parses the packet which last matched
   * 
   * @return the parsed object.
   */
  public Object parse() throws IllegalArgumentException
  {
    if (_lastPacket != null)
      return parse(_lastPacket);
    else
      return null;
  }

  /**
   * Serialization of particleTuples.
   * 
   * The associated parser is actually asked to perform the serialization of the
   * object to a short[] array. This short[] will then be appended to a tuple
   * which matches this Template.
   * 
   * @param obj
   *          the object to be serialized.
   * @param pck
   *          the pck to be modified, if null a pck is created.
   * 
   * @return the pck containing the serialized object.
   */
  public ParticlePacket packet(Object obj, ParticlePacket pck)
      throws IllegalArgumentException
  {
    if (pck == null) pck = new ParticlePacket();

    ParticleTuple ctup = getTuple(pck, _occurences, _payload);
    short[] data = new short[0];

    if (ctup == null)
    {
      // Add the payload before the packet.
      data = _combineArrays(getParser().serialize(_payload), getParser()
          .serialize(obj));
    }
    else
    {
      data = _combineArrays(getParser().serialize(obj), ctup.toArray());
      pck.remove(ctup);
    }

    pck.aclAdd(_tuple.getAclType(), data, 0);

    assert (getTuple(pck, _occurences, _payload) != null);
    return pck;
  }

  protected short[] _combineArrays(short[] src1, short[] src2)
  {
    int cont_len = src1.length + src2.length;
    short[] cont = new short[cont_len]; // a container holding cdata and ndata

    for (int i = 0; i < src1.length; i++)
      cont[i] = src1[i];

    for (int i = 0; i < src2.length; i++)
      cont[src1.length + i] = src2[i];

    return cont;
  }

  protected LinkedList<ParticleTemplate> _templates = new LinkedList<ParticleTemplate>();

  /**
   * add a TupleTemplate to this Template.
   * 
   * @param tupleTemplate
   *          the template to add.
   */
  public void addChild(ParticleTemplate tupleTemplate)
  {
    _templates.add(tupleTemplate);
  }

  public LinkedList<ParticleTemplate> getChilds()
  {
    return _templates;
  }

  /** @} */

  /**
   * @name ActionListener. @{
   */
  private LinkedList<ActionListener> _listeners = new LinkedList<ActionListener>();

  public void addActionListener(ActionListener action)
  {
    _listeners.add(action);
  }

  public void removeActionListener(ActionListener action)
  {
    _listeners.remove(action);
  }

  public List<ActionListener> getActionListeners()
  {
    return _listeners;
  }

  protected void notifyActionListeners()
  {
    ActionEvent event = new ActionEvent(this, 0, MATCH_CMD);
    for (ActionListener listener : _listeners)
    {
      listener.actionPerformed(event);
    }
  }

  /** @} */

  protected Object _userObject = null;

  public void setUserObject(Object o)
  {
    _userObject = o;
  }

  public Object getUserObject()
  {
    return _userObject;
  }

  public Date getTimeStamp()
  {
    return new Date(System.currentTimeMillis());
  }

  public String toString()
  {
    return this.getClass() + " " + _tuple.getAclType() + " " + _payload + " "
        + _occurences;
  }
}
