/*
 * Created on 03.05.2005
 *
 * Copyright (c) 2004
 * Telecooperation Office (TecO), Universitaet Karlsruhe (TH), Germany.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the Universitaet Karlsruhe (TH) nor the names
 *    of its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package edu.teco.particles.pComponent.sender;

import java.text.ParseException;
import java.util.Arrays;
import java.util.LinkedList;

import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;

import edu.teco.particles.libParticle.ParticlePacket;
import edu.teco.particles.libParticle.ParticleSocket;
import edu.teco.particles.libParticle.ParticleSrcId;
import edu.teco.particles.libParticle.ParticleTuple;

/**
 * a proxy class representing a ParticlePacket. Serves as the model for this
 * component.
 * 
 * Functionality: - representing the packet as a tableModel - support for
 * changing properties of packet (like removing tuples)
 * 
 * Encapsulating Packet Properties: - sender ID - receiver ID - tuple data
 * 
 */
public class ProxyParticlePacket implements TableModel
{

  /**
   * Property definition for the proxied ParticlePacket @{
   */

  /*
   * ========================================================================
   * the _real_ ParticlePacket which is "proxied"
   * ======================================================================
   */

  /**
   * @param packet
   */
  public ProxyParticlePacket(ParticlePacket packet)
  {
    setPacket(packet);
  }

  /**
   */
  public ProxyParticlePacket()
  {
  }

  public void setPacket(ParticlePacket packet)
  {
    _tupleList.clear();
    ParticleTuple tuple = packet.firstAcl();
    while (packet.nextAcl(tuple)!=null)
    {
      _tupleList.add(new ProxyParticleTuple(tuple));
      tuple = packet.nextAcl(tuple);
    }
    notifyListeners();
  }

  /** creates the particlePacket currently represented by the TableModel */
  public ParticlePacket getPacket()
  {
	ParticlePacket packet = new ParticlePacket();
	for (ProxyParticleTuple tuple : _tupleList)
		packet.add(tuple.getTuple(), 0);
	
	packet.setSrcId(getSenderId());
	packet.setSeq(getSeqNr());
	
    return packet;
  }

  /*
   * ========================================================================
   * the sender ID of the packet
   * ======================================================================
   */
  /** reference for the sender ID */
  protected ParticleSrcId _sender;

  public void setSenderId(ParticleSrcId sender)
  {
    assert (sender != null) : "trying to set a null sender ID";
    _sender = sender;
  }

  public ParticleSrcId getSenderId()
  {
	if (_sender == null)
		_sender = new ParticleSocket(5555).getID();
    return _sender;
  }

  /*
   * ========================================================================
   * the destination ID of the packet
   * ======================================================================
   */
  /** reference for the destination ID */
  protected ParticleSrcId _destination;

  public void setDestinationId(ParticleSrcId receiver)
  {
    assert (receiver != null) : "trying to set a null receiver ID";
    _destination = receiver;
  }

  public ParticleSrcId getDestinationId()
  {
    return _destination;
  }

  /*
   * ========================================================================
   * the Seq Number of the packet
   * ======================================================================
   */
  /** to the seq Nr */
  protected short _seqNr;

  /**
   * the argument is modded to 255 :) nr % 256.
   * 
   * @param nr
   *          the seqNr for the packet. values larger than 255 will be
   *          truncated.
   */
  public void setSeqNr(short nr)
  {
    if (nr < 0)
    {
      nr = 255;
    }
    ;
    _seqNr = (short) (nr % 256);
    assert (_seqNr >= 0 && _seqNr <= 255);

    notifyListeners();
  }

  public short getSeqNr()
  {
    return _seqNr;
  }

  /** @} */

  /*
   * ========================================================================
   * Table model part:
   * 
   * ------------------------------------- | tuple type | tuple data |
   * ------------------------------------- | sgx | 1 2 2 2 2 | | sgy | 1 2 3 4 4
   * 4 | -------------------------------------
   * 
   * consists of a list of ParticleTuples displaying the type and data
   * properties.
   * ========================================================================
   */

  protected LinkedList<ProxyParticleTuple> _tupleList = new LinkedList<ProxyParticleTuple>();

  protected String[]                  _colDesc   = { "tuple type", "tuple data" };

  protected Class[]                   _colClass  = { String.class, String.class };

  public int getRowCount()
  {
    return _tupleList.size();
  }

  public int getColumnCount()
  {
    return _colDesc.length;
  }

  public String getColumnName(int columnIndex)
  {
    assert (columnIndex < getColumnCount()) : "Index out of range";
    return _colDesc[columnIndex];
  }

  public Class<?> getColumnClass(int columnIndex)
  {
    assert (columnIndex < getColumnCount()) : "Index out of range";
    return _colClass[columnIndex];
  }

  public boolean isCellEditable(int rowIndex, int columnIndex)
  {
    return true;
  }

  public Object getValueAt(int rowIndex, int columnIndex)
  {
    assert (rowIndex < _tupleList.size()) : "Index out of range";
    assert (columnIndex < _colDesc.length) : "Index out of range";

    ProxyParticleTuple row = _tupleList.get(rowIndex);
    String retValue = null;

    switch (columnIndex) {
    case 0:
      retValue = row.getType();
      break;
    case 1:
      retValue = row.getData();
      break;
    }

    return retValue;
  }

  public void setValueAt(Object aValue, int rowIndex, int columnIndex) 
  {
    ProxyParticleTuple row = _tupleList.get(rowIndex);
    switch (columnIndex)
    {
    case 0:
      row.setType((String) aValue);
      break;
    case 1:
      try
      {
        row.setData((String) aValue); 
      }
      catch (Exception e)
      {
        // ignore
      }
      break;
    default:
      assert (false) : "column nummber not known";
    }
  }

  protected LinkedList<TableModelListener> listeners = new LinkedList<TableModelListener>();

  public void addTableModelListener(TableModelListener l)
  {
    listeners.add(l);
  }

  public void removeTableModelListener(TableModelListener l)
  {
    listeners.remove(l);
  }

  public void notifyListeners()
  {
    TableModelEvent e = new TableModelEvent(this);
    for (TableModelListener list : listeners)
    {
      list.tableChanged(e);
    }
  }

  /*
   * ========================================================================
   * support for changing the tuple list
   * =========================================================================
   */

  /**
   * @param is indexes which should be removed.
   */
  public void remove(int[] is)
  {
    Arrays.sort(is);
    for (int i=is.length; i>0; i--)
      _tupleList.remove(is[i-1]);
    notifyListeners();
  }

  /**
   * @param tuple
   */
  public void add(ParticleTuple tuple)
  {
    assert (tuple != null);
    _tupleList.add(new ProxyParticleTuple(tuple));
    notifyListeners();
  }

  public boolean equals(Object arg0)
  {
    if (arg0 instanceof ProxyParticlePacket)
    {
      ParticlePacket pck = ((ProxyParticlePacket) arg0).getPacket();
      return this.equals(pck);
    }
    else if (arg0 instanceof ParticlePacket)
    {
      ParticlePacket other;
      other = (ParticlePacket) arg0;

      if (!other.getSrcId().equals(other.getSrcId())) return false;

      return true;
    }

    return false;
  }

  public int hashCode()
  {
    return getPacket().getSrcId().hashCode();
  }

  public String toString()
  {
    return getPacket().getSrcId().toString();
  }
}
